/*
 * Soubor:  posfix-kalkul.c
 * Datum:   17.03.2011
 * Autor:   Marek Nožka, nozka <@t> spseol <d.t> cz
 * Licence: GNU/GPL 
 * Úloha:   Kalkulátor s reverzní polskou notací
 ****************************************************/
#define _ISOC99_SOURCE
#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdbool.h>
#include <math.h>

/* ************************************************ */
/*#include "easystack.h"                                */
#include "dynmstack.h"

/* ***************    Makra    ******************** */
#define CMDLEN 128
#define VARIABLEC ('z'-'a'+1)

/* *********** Prototypy  funkcí ****************** */
typedef double (*OPER2PAR) (double, double);
typedef double (*OPER1PAR) (double);

void printHelp(void);
double rad2rad(double rad);
double deg2rad(double deg);
double rad2deg(double rad);
void make2wrap(OPER2PAR function, char operationName[]);
void make1wrap(OPER1PAR function, char operationName[]);
double makePlus(double x, double y);
double makeMinus(double x, double y);
double makeTimes(double x, double y);
double makeDivid(double x, double y);
double makeExp(double x);


/* *********** Globální proměnné ****************** */
double (*torad) (double) = deg2rad;     // odkaz na funkci, která převádí na radiány
double (*radto) (double) = rad2deg;     // odkaz na funkci, která převádí na radiány

/****************************************************
 *               Hlavní program.
 ****************************************************/
int main(void)
{
    double number;              // číslo se kterým se má počítat
    char enter[CMDLEN];         // vstup od uživatele
    char message[MSGLEN];
    double variables[VARIABLEC] = { 0 };

    /* Čte řádky až do konce souboru */
    while (scanf("%s", enter) != EOF) {
        if (sscanf(enter, "%lg", &number) == 1) {
            push(number);
        } else {
            if (strcmp(enter, "+") == 0) {
                make2wrap(makePlus, enter);
            } else if (strcmp(enter, "-") == 0) {
                make2wrap(makeMinus, enter);
            } else if (strcmp(enter, "*") == 0) {
                make2wrap(makeTimes, enter);
            } else if (strcmp(enter, "/") == 0) {
                make2wrap(makeDivid, enter);
            } else if (strcmp(enter, "**") == 0) {
                /* n-tá mocnina */
                make2wrap(pow, enter);
            } else if (strcmp(enter, "sqrt") == 0
                       || strcmp(enter, "q") == 0) {
                make1wrap(sqrt, enter);
            } else if (strcmp(enter, "exp") == 0
                       || strcmp(enter, "x") == 0) {
                /* exponenciální funkce */
                make1wrap(makeExp, enter);
            } else if (strcmp(enter, "ln") == 0) {
                make1wrap(log, enter);
            } else if (strcmp(enter, "log") == 0) {
                make1wrap(log10, enter);
                /* Goniometrické funkce */
            } else if (strcmp(enter, "r") == 0) {
                /* přepne kalkulačku do radiánů */
                torad = rad2rad;
                radto = rad2rad;
            } else if (strcmp(enter, "d") == 0) {
                /* přepne kalkulačku do stupňů */
                torad = deg2rad;
                radto = rad2deg;
            } else if (strcmp(enter, "sin") == 0) {
                make1wrap(sin, enter);
            } else if (strcmp(enter, "cos") == 0) {
                make1wrap(cos, enter);
            } else if (strcmp(enter, "tan") == 0
                       || strcmp(enter, "tg") == 0) {
                make1wrap(tan, enter);
            } else if (strcmp(enter, "asin") == 0) {
                make1wrap(asin, enter);
            } else if (strcmp(enter, "acos") == 0) {
                make1wrap(acos, enter);
            } else if (strcmp(enter, "atan") == 0
                       || strcmp(enter, "atg") == 0) {
                make1wrap(atan, enter);
            } else if (enter[0] == '>' && enter[1] >= 'a'
                       && enter[1] <= 'z') {
                /* ukládá do proměnné */
                variables[enter[1] - 'a'] = pop();
                push(variables[enter[1] - 'a']);
            } else if (enter[0] == '<' && enter[1] >= 'a'
                       && enter[1] <= 'z') {
                /* vybírá obsah proměnné */
                push(variables[enter[1] - 'a']);
            } else if (strcmp(enter, "rem") == 0
                       || strcmp(enter, "m") == 0) {
                pop();
                spstack(message);       // zjistím, co je v zásobníku
                printf(">>> %s\n", message);    // ... a vytisknu to 
            } else if (strcmp(enter, "clr") == 0
                       || strcmp(enter, "c") == 0) {
                stackclear();   // vymaže celý zásobník
            } else if (strcmp(enter, "p") == 0) {       // jen tiskne zásobník
                spstack(message);       // zjistím, co je v zásobníku
                printf(">>> %s\n", message);    // ... a vytisknu to 
            } else if (strcmp(enter, "help") == 0
                       || strcmp(enter, "h") == 0) {
                printHelp();
            } else if (strcmp(enter, "end") == 0
                       || strcmp(enter, "e") == 0) {
                return 0;
            } else {
                fprintf(stderr, " error > %s\n", enter);
                fprintf(stderr, " error > Neplatný vstup\n");
                spstack(message);       // zjistím, co je v zásobníku
                printf(">>> %s\n", message);    // ... a vytisknu to 
            }
        }


    }
    stackclear();
    return 0;
}

/****************************************************
 *                 Funkce
 ****************************************************/

/* Tiskne malou nápovědu */
void printHelp(void)
{
    printf
        ("Kalkulátor s polskou reverzní notací. Pořadí matematických\n"
         "operací je dáno přímo zápisem. Například: \n\n"
         "(5 * (4 + 3) ) / (10-3) odpovídá 5 4 3 + * 10 3 - /\n"
         "\nPodporované matematické operace a funkce:\n\n" "+ - * /\n"
         "**       mocnina\n" "sqrt|q   druhá odmocnina\n"
         "sin cos tg|tan\n" "asin acos atg|atan\n" "exp|x ln log\n"
         "\nPříkazy:\n"
         "d        přepne pro počítání ve stupních\n"
         "r        přepne pro počítání v radiánech\n"
         "rem|m    vymaže poslední číslo ze zásobníku\n"
         "clr|c    vymaže celý zásobník\n"
         "end|e    ukončí program\n"
         "\nPaměťové registry/proměnné a-z:\n"
         ">a       uloží poslední číslo v zásobníku do registru a \n"
         "<a       uloží poslední číslo v registru a do zásobníku \n"
         "         pro jména registrů se mohou použít písmena a-z\n"
         "\n");
}

/* Funkce vyzvedne ze zásobníku 2 čísla, provede s nimi
 * matematickou operace danou parametrem function a výsledek
 * uloží zpět do zásobníku.
 * Funkce vrací FALSE pokud operaci nelze provést.
 */
bool make2param(OPER2PAR function)
{
    double x, y;
    if (isnan(y = pop())) {
        return false;
    }
    if (isnan(x = pop())) {
        push(y);
        return false;
    }
    push(function(x, y));
    return true;
}

/* Funkce vyzvedne ze zásobníku číslo, provede s nimi
 * matematickou operace danou parametrem function a výsledek
 * uloží zpět do zásobníku.
 * Funkce vrací FALSE pokud operaci nelze provést.
 */
bool make1param(OPER1PAR function)
{
    double x;
    if (isnan(x = pop())) {
        return false;
    }
    if (function == sin || function == cos || function == tan) {
        push(function(torad(x)));
    } else if (function == asin || function == acos || function == atan) {
        push(radto(function(x)));
    } else {
        push(function(x));
    }
    return true;
}

/* Tato funkce jen vhodně obaluje make2param a komunikuje 
 * s uživatelem*/
void make2wrap(OPER2PAR function, char operationName[])
{
    char message[MSGLEN];
    if (!make2param(function)) {
        fprintf(stderr, " error > %s vyžaduje dva parametry\n",
                operationName);
    }
    spstack(message);           // zjistím, co je v zásobníku
    printf(">>> %s\n", message);        // ... a vytisknu to 
}

/* Tato funkce jen vhodně obaluje make1param a komunikuje 
 * s uživatelem*/
void make1wrap(OPER1PAR function, char operationName[])
{
    char message[MSGLEN];
    if (!make1param(function)) {
        fprintf(stderr, " error > %s vyžaduje jeden parametr\n",
                operationName);
    }
    spstack(message);           // zjistím, co je v zásobníku
    printf(">>> %s\n", message);        // ... a vytisknu to 

}

/*    plus      */
double makePlus(double x, double y)
{
    return x + y;
}

/*    minus      */
double makeMinus(double x, double y)
{
    return x - y;
}

/*    krát      */
double makeTimes(double x, double y)
{
    return x * y;
}

/*    děleno      */
double makeDivid(double x, double y)
{
    return x / y;
}

/*    exp(x)      */
double makeExp(double x)
{
    double sum = 1.0;           // součet všech členů
    double element = 1.0;       // člen posloupnosti
    double i = 1.0;             // pořadí členu

    while (element > 1E-8) {
        element *= x / i;
        sum += element;
        i++;
    }
    return sum;
}

/* Funkce převádí stupně na radiány */
double deg2rad(double deg)
{
    return M_PI * deg / 180;
}

double rad2rad(double rad)
{
    return rad;
}

double rad2deg(double deg)
{
    return 180 * deg / M_PI;
}
